'''Unittests for dataset class.'''

import unittest
import h5py
import numpy as np
import pandas as pd
import ensembleset.dataset as ds
import ensembleset.feature_methods as fm

# pylint: disable=protected-access

class TestDataSetInit(unittest.TestCase):
    '''Tests for main data set generator class initialization.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df_without_strings = pd.DataFrame({
            'feature1': [0,1],
            'feature2': [3,4],
            'feature3': [5,6]
        })

        self.dataset_without_string_feature = ds.DataSet(
            'feature2',
            self.dummy_df_without_strings,
            test_data=self.dummy_df_without_strings
        )

        self.dummy_df_with_strings = pd.DataFrame({
            'feature1': [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset_with_string_feature = ds.DataSet(
            'feature2',
            self.dummy_df_with_strings,
            test_data=self.dummy_df_with_strings,
            string_features=['feature3']
        )


    def test_class_arguments(self):
        '''Tests assignments of class attributes from user arguments.'''

        self.assertTrue(isinstance(self.dataset_with_string_feature.label, str))
        self.assertTrue(isinstance(self.dataset_with_string_feature.train_data, pd.DataFrame))
        self.assertTrue(isinstance(self.dataset_with_string_feature.test_data, pd.DataFrame))
        self.assertTrue(isinstance(self.dataset_with_string_feature.string_features, list))
        self.assertEqual(self.dataset_with_string_feature.string_features[0], 'feature3')

        with self.assertRaises(TypeError):
            ds.DataSet(
                2,
                self.dummy_df_with_strings,
                test_data=self.dummy_df_with_strings,
                string_features=['feature3']
            )

        with self.assertRaises(TypeError):
            ds.DataSet(
                'feature2',
                'Not a Pandas Dataframe',
                test_data=self.dummy_df_with_strings,
                string_features=['feature3']
            )

        with self.assertRaises(TypeError):
            ds.DataSet(
                'feature2',
                self.dummy_df_with_strings,
                test_data='Not a Pandas Dataframe',
                string_features=['feature3']
            )

        with self.assertRaises(TypeError):
            ds.DataSet(
                'feature2',
                self.dummy_df_with_strings,
                test_data=self.dummy_df_with_strings,
                string_features='Not a list of features'
            )


    def test_output_creation(self):
        '''Tests the creation of the HDF5 output sink.'''

        hdf = h5py.File('data/dataset.hdf5', 'a')

        self.assertTrue('train' in hdf)
        self.assertTrue('test' in hdf)


    def test_pipeline_options(self):
        '''Tests the creation of feature engineering pipeline options'''

        self.assertTrue(isinstance(self.dataset_with_string_feature.string_encodings, dict))
        self.assertTrue(isinstance(self.dataset_with_string_feature.engineerings, dict))


class TestDataPipelineGen(unittest.TestCase):
    '''Tests for data pipeline generator function.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = pd.DataFrame({
            'feature1': [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset = ds.DataSet(
            'feature2',
            self.dummy_df,
            test_data=self.dummy_df,
            string_features=['feature3']
        )


    def test_generate_data_pipeline(self):
        '''Tests the data pipeline generation function.'''

        pipeline=self.dataset._generate_data_pipeline(2)

        self.assertEqual(len(pipeline), 3)

        for operation, parameters in pipeline.items():
            self.assertTrue(isinstance(operation, str))
            self.assertTrue(isinstance(parameters, dict))


class TestFeatureSelection(unittest.TestCase):
    '''Tests for data pipeline generator function.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = pd.DataFrame({
            1: [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset = ds.DataSet(
            'feature2',
            self.dummy_df,
            test_data=self.dummy_df,
            string_features=['feature3']
        )


    def test_select_features(self):
        '''Tests feature selection function.'''

        features=self.dataset._select_features(3, self.dummy_df)

        self.assertEqual(len(features), 2)

        for feature in features:
            self.assertTrue(isinstance(feature, str))

        self.assertFalse('feature2' in features)


class TestDatasetGeneration(unittest.TestCase):
    '''Tests dataset generation.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = pd.DataFrame({
            1: [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset = ds.DataSet(
            'feature2',
            self.dummy_df,
            test_data=self.dummy_df,
            string_features=['feature3']
        )

        self.dataset.make_datasets(2, 2, 1)


    def test_make_datasets(self):
        '''Tests generation of datasets.'''

        hdf = h5py.File('data/dataset.hdf5', 'a')

        training_datasets=hdf['train']
        self.assertEqual(len(training_datasets), 2)

        testing_datasets=hdf['test']
        self.assertEqual(len(testing_datasets), 2)


class TestFeatureMethods(unittest.TestCase):
    '''Tests feature engineering method functions.'''

    def setUp(self):
        '''Dummy DataFrames for tests.'''

        self.dummy_df = pd.DataFrame({
            'feature1': [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })


    def test_onehot_encoding(self):
        '''Tests string features onehot encoder.'''

        train_df, test_df=fm.onehot_encoding(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature3'],
            {'sparse_output': False}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_ordinal_encoding(self):
        '''Tests string feature ordinal encoder.'''

        train_df, test_df=fm.ordinal_encoding(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature3'],
            {
                'handle_unknown': 'use_encoded_value',
                'unknown_value': np.nan  
            }
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_poly_features(self):
        '''Tests polynomial feature transformer.'''

        train_df, test_df=fm.poly_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature2'],
            {'degree': 2}

        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_spline_features(self):
        '''Tests spline features transformer.'''

        train_df, test_df=fm.spline_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature2'],
            {'n_knots': 2}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_log_features(self):
        '''Tests log features transformer.'''

        train_df, test_df=fm.log_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1'],
            {'base': '2'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.log_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1'],
            {'base': 'e'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.log_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1'],
            {'base': '10'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_ratio_features(self):
        '''Tests ratio feature transformer.'''

        train_df, test_df=fm.ratio_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'div_zero_value': np.nan}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_exponential_features(self):
        '''Tests exponential features transformer.'''

        train_df, test_df=fm.exponential_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'base': 'e'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.exponential_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'base': '2'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_sum_features(self):
        '''Tests sum features transformer.'''

        train_df, test_df=fm.sum_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'n_addends': 2}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.sum_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'n_addends': 4}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_difference_features(self):
        '''Tests difference features transformer.'''

        train_df, test_df=fm.difference_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'n_subtrahends': 2}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.difference_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['feature1', 'feature2'],
            {'n_subtrahends': 4}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


'''Generates variations of a dataset using a pool of feature engineering
techniques. Used for training ensemble models.'''

from pathlib import Path
from random import choice, shuffle

import h5py
import numpy as np
import pandas as pd

import ensembleset.feature_methods as fm


class DataSet:
    '''Dataset generator class.'''

    def __init__(
            self,
            label: str,
            train_data: pd.DataFrame,
            test_data: pd.DataFrame = None,
            string_features: list = None
        ):

        # Type check the user arguments and assign them to attributes
        if isinstance(label, str):
            self.label = label

        else:
            raise TypeError('Label is not a string.')

        if isinstance(train_data, pd.DataFrame):
            train_data.columns = train_data.columns.astype(str)
            self.train_data = train_data

        else:
            raise TypeError('Train data is not a Pandas DataFrame.')

        if isinstance(test_data, pd.DataFrame) or test_data is None:
            test_data.columns = test_data.columns.astype(str)
            self.test_data = test_data

        else:
            raise TypeError('Test data is not a Pandas DataFrame.')

        if isinstance(string_features, list) or string_features is None:
            self.string_features = string_features

        else:
            raise TypeError('String features is not a list.')

        # Create the HDF5 output
        Path('data').mkdir(parents=True, exist_ok=True)

        with h5py.File('data/dataset.hdf5', 'a') as hdf:
            _ = hdf.require_group('train')
            _ = hdf.require_group('test')

        # Define the feature engineering pipeline operations
        self.string_encodings={
            'onehot_encoding': {'sparse_output': False},
            'ordinal_encoding': {
                'handle_unknown': 'use_encoded_value',
                'unknown_value': np.nan  
            }
        }

        self.engineerings={
            'poly_features': {
                'degree': [2, 3],
                'interaction_only': [True, False],
            },
            'spline_features': {
                'n_knots': [5],
                'degree': [2, 3, 4],
                'knots': ['uniform', 'quantile'],
                'extrapolation': ['error', 'constant', 'linear', 'continue', 'periodic']
            },
            'log_features': {
                'base': ['2', 'e', '10']
            },
            'ratio_features': {
                'div_zero_value': [np.nan]
            },
            'exponential_features': {
                'base': ['2', 'e']
            },
            'sum_features': {
                'n_addends': [2,3,4]
            },
            'difference_features': {
                'n_subtrahends': [2,3,4]
            }
        }


    def make_datasets(self, n_datasets:int, n_features:int, n_steps:int):
        '''Makes n datasets with different feature subsets and pipelines.'''

        hdf = h5py.File('data/dataset.hdf5', 'w')

        for n in range(n_datasets):

            train_df = self.train_data.copy()
            test_df = self.test_data.copy()
            pipeline = self._generate_data_pipeline(n_steps)

            for operation, arguments in pipeline.items():
                func = getattr(fm, operation)

                if operation in self.string_encodings:
                    train_df, test_df = func(
                        train_df,
                        test_df,
                        self.string_features,
                        arguments
                    )

                else:
                    features = self._select_features(n_features, train_df)

                    train_df, test_df = func(
                        train_df,
                        test_df,
                        features,
                        arguments
                    )

            _ = hdf.create_dataset(f'train/{n}', data=np.array(train_df))
            _ = hdf.create_dataset(f'test/{n}', data=np.array(test_df))

        hdf.close()


    def _select_features(self, n_features:int, data_df:pd.DataFrame):
        '''Selects a random subset of features.'''

        features = data_df.columns.to_list()#.astype(str).to_list()
        features.remove(self.label)
        shuffle(features)
        features = features[:n_features]

        return features


    def _generate_data_pipeline(self, n_steps:int):
        '''Generates one random sequence of feature engineering operations. Starts with
        a string encoding method if we have string features.'''

        pipeline={}

        # Choose a string encoding method, if needed
        if self.string_features is not None:
            options = list(self.string_encodings.keys())
            selection = choice(options)
            pipeline[selection] = self.string_encodings[selection]

        # Construct a random sequence of feature engineering operations
        operations = list(self.engineerings.keys())
        shuffle(operations)
        operations = operations[:n_steps]

        for operation in operations:

            pipeline[operation] = {}
            parameters = self.engineerings[operation]

            for parameter, values in parameters.items():

                value = choice(values)
                pipeline[operation][parameter] = value

        return pipeline




'''Collection of functions to run feature engineering operations.'''

from math import e
from itertools import permutations, combinations
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, OrdinalEncoder, PolynomialFeatures, SplineTransformer


def onehot_encoding(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's one hot encoder.'''

    encoder=OneHotEncoder(**kwargs)

    encoded_data=encoder.fit_transform(train_df[features])
    encoded_df=pd.DataFrame(encoded_data, columns=encoder.get_feature_names_out())
    train_df.drop(features, axis=1, inplace=True)
    train_df=pd.concat([train_df, encoded_df], axis=1)

    if test_df is not None:
        encoded_data=encoder.transform(test_df[features])
        encoded_df=pd.DataFrame(encoded_data, columns=encoder.get_feature_names_out())
        test_df.drop(features, axis=1, inplace=True)
        test_df=pd.concat([test_df, encoded_df], axis=1)

    return train_df, test_df


def ordinal_encoding(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's label encoder.'''

    encoder=OrdinalEncoder(**kwargs)

    train_df[features]=encoder.fit_transform(train_df[features])

    if test_df is not None:
        test_df[features]=encoder.transform(test_df[features])

    return train_df, test_df


def poly_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's polynomial feature transformer..'''

    transformer=PolynomialFeatures(**kwargs)
    imputer=SimpleImputer(strategy='mean')

    imputed_data=imputer.fit_transform(train_df[features])
    encoded_data=transformer.fit_transform(imputed_data)
    encoded_df=pd.DataFrame(encoded_data, columns=transformer.get_feature_names_out())
    train_df.drop(features, axis=1, inplace=True)
    train_df=pd.concat([train_df, encoded_df], axis=1)

    if test_df is not None:

        imputed_data=imputer.transform(test_df[features])
        encoded_data=transformer.transform(imputed_data)
        encoded_df=pd.DataFrame(encoded_data, columns=transformer.get_feature_names_out())
        test_df.drop(features, axis=1, inplace=True)
        test_df=pd.concat([test_df, encoded_df], axis=1)

    return train_df, test_df


def spline_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's polynomial feature transformer..'''

    transformer=SplineTransformer(**kwargs)
    imputer=SimpleImputer(strategy='mean')

    imputed_data=imputer.fit_transform(train_df[features])
    encoded_data=transformer.fit_transform(imputed_data)
    encoded_df=pd.DataFrame(encoded_data, columns=transformer.get_feature_names_out())
    train_df.drop(features, axis=1, inplace=True)
    train_df=pd.concat([train_df, encoded_df], axis=1)

    if test_df is not None:

        imputed_data=imputer.transform(test_df[features])
        encoded_data=transformer.transform(imputed_data)
        encoded_df=pd.DataFrame(encoded_data, columns=transformer.get_feature_names_out())
        test_df.drop(features, axis=1, inplace=True)
        test_df=pd.concat([test_df, encoded_df], axis=1)

    return train_df, test_df


def log_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Takes log of feature, uses sklearn min-max scaler if needed
    to avoid undefined log errors.'''

    for feature in features:
        if min(train_df[feature]) <= 0:

            scaler=MinMaxScaler(feature_range=(1, 100))

            train_df[feature]=scaler.fit_transform(train_df[feature].to_frame())
            test_df[feature]=scaler.transform(test_df[feature].to_frame())

        if kwargs['base'] == '2':
            train_df[f'{feature}_log2']=np.log2(train_df[feature])
            test_df[f'{feature}_log2']=np.log2(test_df[feature])

        if kwargs['base'] == 'e':
            train_df[f'{feature}_ln']=np.log(train_df[feature])
            test_df[f'{feature}_ln']=np.log(test_df[feature])

        if kwargs['base'] == '10':
            train_df[f'{feature}_log10']=np.log10(train_df[feature])
            test_df[f'{feature}_log10']=np.log10(test_df[feature])

    return train_df, test_df


def ratio_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds every possible ratio feature, replaces divide by zero errors
    with np.nan.'''

    feature_pairs=permutations(features, 2)

    for feature_a, feature_b in feature_pairs:

        quotient = np.divide(
            np.array(train_df[feature_a]),
            np.array(train_df[feature_b]),
            out=np.array([kwargs['div_zero_value']]*len(train_df[feature_a])),
            where=np.array(train_df[feature_b]) != 0
        )

        train_df[f'{feature_a}_over_{feature_b}'] = quotient

        quotient = np.divide(
            np.array(test_df[feature_a]),
            np.array(test_df[feature_b]),
            out=np.array([kwargs['div_zero_value']]*len(test_df[feature_a])),
            where=np.array(test_df[feature_b]) != 0
        )

        test_df[f'{feature_a}_over_{feature_b}'] = quotient

    return train_df, test_df


def exponential_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds exponential features with base 2 or base e.'''

    for feature in features:
        if kwargs['base'] == 'e':
            train_df[f'{feature}_exp_base_e'] = e**train_df[feature]
            test_df[f'{feature}_exp_base_e'] = e**test_df[feature]

        elif kwargs['base'] == '2':
            train_df[f'{feature}_exp_base_e'] = 2**train_df[feature]
            test_df[f'{feature}_exp_base_e'] = 2**test_df[feature]

    return train_df, test_df


def sum_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds sum features for variable number of addends.'''

    if kwargs['n_addends'] > len(features):
        n_addends=len(features)

    else:
        n_addends=kwargs['n_addends']

    addend_sets=combinations(features, n_addends)

    for i, addend_set in enumerate(addend_sets):

        train_sum = [0]*len(train_df)
        test_sum = [0]*len(test_df)

        for addend in addend_set:

            train_sum += train_df[addend]
            test_sum += test_df[addend]

        train_df[f'sum_feature_{i}'] = train_sum
        test_df[f'sum_feature_{i}'] = test_sum

    return train_df, test_df


def difference_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds difference features for variable number of subtrahends.'''

    if kwargs['n_subtrahends'] > len(features):
        n_subtrahends=len(features)

    else:
        n_subtrahends=kwargs['n_subtrahends']

    subtrahend_sets=combinations(features, n_subtrahends)

    for i, subtrahend_set in enumerate(subtrahend_sets):

        train_difference = train_df[subtrahend_set[0]]
        test_difference = test_df[subtrahend_set[0]]

        for subtrahend in subtrahend_set[1:]:

            train_difference -= train_df[subtrahend]
            test_difference -= test_df[subtrahend]

        train_df[f'difference_feature_{i}'] = train_difference
        test_df[f'difference_feature_{i}'] = test_difference

    return train_df, test_df